!----------------------------------------------------------
!   resolution de les quations d'euler 2d 
!   - sur grille structure curviligne
!   - domaine topologiquement equivalent a un carre : 
!             NORD
!            _______  
!            |     |  ^j
!    OUEST   |     |  |     EST
!            _______  --> i

!              SUD
!   - schema volumes finis, decentre d'ordre 1, explicite en temps
!   - fichiers de donnees : maillage.data
!----------------------------------------------------------

program euler2d

  !----------- sous programmes ------------------------------------ 
  use mod_precision   !- (double ou simple precision)
  use mod_flux
  use mod_gauss
  !----------------------------------------------------------------

  implicit none

  !------------ variables -----------------------------------------
  integer :: i,j,imax,jmax,nmax,n,k,pas_affichage,i_entree
  character(len=100) :: fich,numero
  real(pr) :: L,dx,dy,cfl,invdt,dt,vmax,tmax
  real(pr) :: r_gaz,norme_u,perimetre,ut,un
  real(pr) :: adim_aire,adim_flux_masse,adim_flux_qmvt,adim_flux_e,adim_t
  real(pr) :: adim_longueur,residu_tot,residu_init
  real(pr) :: rspec
  real(pr) :: gamma,gamma1,gamma2,uc,h,nx,ny,uij,vij,c,ec
  real(pr), dimension(1:4) :: residu
  real(pr), dimension(1:4,1:4) :: mat_a,mat_id
  real(pr), dimension(:,:), allocatable :: x,y,xc,yc
  real(pr), dimension(:,:), allocatable :: rho,ux,uy,t,e,p,a,b
  real(pr), dimension(:,:), allocatable :: aire,taux,tauy,nux,nuy,l_tau,l_nu
  real(pr), dimension(:,:,:), allocatable :: vect_u,vect_unew,flux_tau,flux_nu,res
  real(pr), dimension(:,:,:), allocatable :: du1,du2
  real(pr), dimension(:,:,:,:), allocatable :: a_taup,a_taum,a_nup,a_num,d
  !----------------------------------------------------------------


  !------------- programme ----------------------------------------

  !--- lecture du maillage
  open(unit=10,file='maillage.data')

  !-nb de points dans la grille
  read(10,*) imax,jmax   

  !-nb de mailles
  imax=imax-1
  jmax=jmax-1   

  allocate(x(0:imax,0:jmax),y(0:imax,0:jmax))
  do i=0,imax
     do j=0,jmax
        read(10,*) x(i,j),y(i,j)
     enddo
  enddo
  close(10)

  !--- calcul des aires des mailles
  allocate(aire(1:imax,1:jmax))
  do i=1,imax
     do j=1,jmax
        aire(i,j)=0.5_pr*( (x(i,j)-x(i-1,j-1))*(y(i-1,j)-y(i,j-1))             &
             &            -(y(i,j)-y(i-1,j-1))*(x(i-1,j)-x(i,j-1)) )
     enddo
  enddo

  !--- calcul des normales unitaires et des longeurs des interfaces
  allocate(taux(0:imax,1:jmax),tauy(0:imax,1:jmax),l_tau(0:imax,1:jmax))
  allocate(nux(1:imax,0:jmax),nuy(1:imax,0:jmax),l_nu(1:imax,0:jmax))
  do i=0,imax
     do j=1,jmax
        taux(i,j) = y(i,j)-y(i,j-1)
        tauy(i,j) = -(x(i,j)-x(i,j-1))
        l_tau(i,j)=sqrt(taux(i,j)**2+tauy(i,j)**2)
        taux(i,j)=taux(i,j)/l_tau(i,j)
        tauy(i,j)=tauy(i,j)/l_tau(i,j)
     end do
  end do
  do i=1,imax
     do j=0,jmax
        nux(i,j) = -(y(i,j)-y(i-1,j))
        nuy(i,j) = x(i,j)-x(i-1,j)
        l_nu(i,j)=sqrt(nux(i,j)**2+nuy(i,j)**2)
        nux(i,j)=nux(i,j)/l_nu(i,j)
        nuy(i,j)=nuy(i,j)/l_nu(i,j)
     end do
  end do

  !--- donne initiale : 
  allocate(rho(0:imax+1,0:jmax+1),ux(0:imax+1,0:jmax+1),uy(0:imax+1,0:jmax+1),t(0:imax+1,0:jmax+1))
  do i=1,imax
     do j=1,jmax

        rho(i,j)=4.528e-7_pr 
        ux(i,j) =2075._pr
        uy(i,j) =1198._pr
        t(i,j)  =249.0198

     end do
  end do

  !--- conditions aux limites : 
  !-- exemple 1 : 
  rho(0,:)=rho(1,:)      !- ouest (sortie)
  ux(0,:)=ux(1,:)
  uy(0,:)=uy(1,:)
  t(0,:)=t(1,:)

  rho(imax+1,:)=rho(imax,:)    !- est (sortie)
  ux(imax+1,:)=ux(imax,:)
  uy(imax+1,:)=uy(imax,:)
  t(imax+1,:)=t(imax,:)

  rho(:,0)=rho(:,1)          !- sud (paroi solide)
  t(:,0)=t(:,1)
  do i=1,imax
     un=nux(i,0)*ux(i,1)+nuy(i,0)*uy(i,1)
     ut=-nuy(i,0)*ux(i,1)+nux(i,0)*uy(i,1)
     ux(i,0)=nux(i,0)*(-un)-nuy(i,0)*ut
     uy(i,0)=nuy(i,0)*(-un)+nux(i,0)*ut
  end do

  rho(1:imax,jmax+1)=4.528e-7_pr     !- nord (entre)
  ux(1:imax,jmax+1)=2075._pr
  uy(1:imax,jmax+1)=1198._pr
  t(1:imax,jmax+1)=249.0198

  !--- valeurs de t aux 4 mailles fictives de coin : inutilises dans le schma
  !--- mais donne division par 0 dans le calcul de a
  t(0,0)=1._pr
  t(0,jmax+1)=1._pr
  t(imax+1,jmax+1)=1._pr
  t(imax+1,0)=1._pr

  !--- calcul des quantits drives
  allocate(p(0:imax+1,0:jmax+1),a(0:imax+1,0:jmax+1),b(0:imax+1,0:jmax+1),e(0:imax+1,0:jmax+1))
  r_gaz=1.3806503e-23_pr/.663e-25_pr
  p=rho*r_gaz*t
  b=sqrt(3*r_gaz*t)
  a=rho/(8_pr*b**3)
  e=0.5_pr*rho*(ux**2+uy**2)+3._pr/2._pr*rho*r_gaz*t

  !--- calcul du vecteur des quantits conservatives
  allocate(vect_u(4,1:imax,1:jmax),vect_unew(4,1:imax,1:jmax))
  vect_u(1,:,:)=rho(1:imax,1:jmax)
  vect_u(2,:,:)=rho(1:imax,1:jmax)*ux(1:imax,1:jmax)
  vect_u(3,:,:)=rho(1:imax,1:jmax)*uy(1:imax,1:jmax)
  vect_u(4,:,:)=e(1:imax,1:jmax)
  vect_unew=vect_u

  !--- adimensionneurs
  adim_flux_masse=sqrt(sum(((rho(1:imax,1:jmax)*ux(1:imax,1:jmax))**2               &
       &                   +(rho(1:imax,1:jmax)*uy(1:imax,1:jmax))**2)*aire))/sum(aire)
  adim_flux_qmvt=sqrt(sum(((rho(1:imax,1:jmax)*ux(1:imax,1:jmax)**2)**2                 &
       &                           +2._pr*(rho(1:imax,1:jmax)*ux(1:imax,1:jmax)*uy(1:imax,1:jmax))**2    & 
       &                           +(rho(1:imax,1:jmax)*uy(1:imax,1:jmax)**2)**2                         &
       &                           +2._pr*p(1:imax,1:jmax)**2)*aire))/sum(aire)
  adim_flux_e=sqrt(sum((((e(1:imax,1:jmax)+p(1:imax,1:jmax))*ux(1:imax,1:jmax))**2                &
       &                +((e(1:imax,1:jmax)+p(1:imax,1:jmax))*uy(1:imax,1:jmax))**2)*aire))/sum(aire)
  adim_t=sqrt((x(0,jmax)-x(0,0))**2+(y(0,jmax)-y(0,0))**2)/2000._pr
  adim_longueur=sqrt((x(0,jmax)-x(0,0))**2+(y(0,jmax)-y(0,0))**2)
  adim_aire=sum(aire)
  allocate(Res(1:4,1:imax,1:jmax))
  residu_tot=1._pr

  open(unit=112,file='SORTIES/residu.dat')

  !--- allocation des flux
  allocate(flux_tau(4,0:imax,1:jmax),flux_nu(4,1:imax,0:jmax))

  !--- matrices et vecteurs pour l'implicite
  mat_id=0._pr
  mat_id(1,1)=1._pr
  mat_id(2,2)=1._pr
  mat_id(3,3)=1._pr
  mat_id(4,4)=1._pr
  allocate(d(4,4,1:imax,1:jmax))
  allocate(a_taup(4,4,0:imax,1:jmax),a_taum(4,4,0:imax,1:jmax))
  a_taup=0._pr ; a_taum=0._pr
  allocate(a_nup(4,4,1:imax,0:jmax),a_num(4,4,1:imax,0:jmax))
  a_nup=0._pr ; a_num=0._pr
  allocate(du1(4,0:imax,0:jmax))
  du1=0._pr
  allocate(du2(4,1:imax+1,1:jmax+1))
  du2=0._pr

  !--- boucle schema implicite
  n=1
  boucle_temps:do 
     !-- calcul des flux
     do i=0,imax
        do j=1,jmax
           flux_tau(:,i,j) = l_tau(i,j)*(fluxp(rho(i,j),ux(i,j),uy(i,j),e(i,j),p(i,j),t(i,j)               &
                ,a(i,j),b(i,j),taux(i,j),tauy(i,j))                         &
                &                      + fluxm(rho(i+1,j),ux(i+1,j),uy(i+1,j),e(i+1,j),p(i+1,j),t(i+1,j)   &
                &                              ,a(i+1,j),b(i+1,j),taux(i,j),tauy(i,j)) ) 
        end do
     end do

     do i=1,imax
        do j=0,jmax
           flux_nu(:,i,j) = l_nu(i,j)*(fluxp(rho(i,j),ux(i,j),uy(i,j),e(i,j),p(i,j),t(i,j)                 &
                &                            ,a(i,j),b(i,j),nux(i,j),nuy(i,j))                             &
                &                      + fluxm(rho(i,j+1),ux(i,j+1),uy(i,j+1),e(i,j+1),p(i,j+1),t(i,j+1)   &
                &                              ,a(i,j+1),b(i,j+1),nux(i,j),nuy(i,j)) ) 
        end do
     end do

     !--- pas de temps (avec monte en cfl)
     invdt=0._pr 
     do i=1,imax
        do j=1,jmax
           norme_u=sqrt(ux(i,j)**2+uy(i,j)**2)
           perimetre=l_tau(i-1,j)+l_tau(i,j)+l_nu(i,j-1)+l_nu(i,j)
           invdt=max(invdt,(norme_u+b(i,j)*perimetre/aire(i,j)))
        end do
     end do


     if (n<100) then
        cfl=0.9_pr
     elseif (n<200) then
        cfl=cfl+0.1_pr
     else
        cfl=10000._pr
     end if

     dt=cfl/invdt

     !-- calcul du second membre
     do i=1,imax
        do j=1,jmax
           Res(:,i,j)=(flux_tau(:,i,j)-flux_tau(:,i-1,j)) + (flux_nu(:,i,j)-flux_nu(:,i,j-1))
        end do
     end do

     !--- calcul rsidu total
     residu=0._pr
     do i=1,imax
        do j=1,jmax
           residu=residu+Res(:,i,j)**2*aire(i,j)
        end do
     end do

     residu(1)=(sqrt(residu(1))/adim_aire)/adim_flux_masse
     residu(2)=(sqrt(residu(2))/adim_aire)/adim_flux_qmvt
     residu(3)=(sqrt(residu(3))/adim_aire)/adim_flux_qmvt
     residu(4)=(sqrt(residu(4))/adim_aire)/adim_flux_e

     residu_tot=sqrt(sum(residu**2))/adim_t*adim_longueur

     if (n==1) then
        residu_init=residu_tot
     end if
     residu_tot=residu_tot/residu_init

     write(112,*) n,residu_tot
     if (mod(n,10)==0.or.n==1) then
        print*,'iteration ',n
        print*,'residu=',residu_tot
     end if

     if (residu_tot<=1e-5_pr) then
        exit boucle_temps
     end if

     !--- calcul des matrices 
     gamma=5._pr/3._pr
     gamma1=gamma-1._pr
     gamma2=gamma-2._pr
     do j=1,jmax
        do i=0,imax
           un=ux(i,j)*taux(i,j)+uy(i,j)*tauy(i,j)
           uij=ux(i,j); vij=uy(i,j)
           nx=taux(i,j)
           ny=tauy(i,j)
           ec=(ux(i,j)**2+uy(i,j)**2)/2._pr
           h=(e(i,j)+p(i,j))/rho(i,j)

           mat_a(1,1)=0._pr; mat_a(1,2)=nx; mat_a(1,3)=ny; mat_a(1,4)=0._pr
           mat_a(2,1)=-uij*un+gamma1*ec*nx; mat_a(2,2)=un-gamma2*uij*nx; mat_a(2,3)=uij*ny-gamma1*vij*nx; mat_a(2,4)=gamma1*nx
           mat_a(3,1)=-vij*un+gamma1*ec*ny; mat_a(3,2)=vij*nx-gamma1*uij*ny; mat_a(3,3)=un-gamma2*vij*ny; mat_a(3,4)=gamma1*ny
           mat_a(4,1)=(gamma1*ec-h)*un; mat_a(4,2)=-gamma1*uij*un+h*nx; mat_a(4,3)=-gamma1*vij*un+h*ny; mat_a(4,4)=gamma*un

           c=sqrt(gamma*r_gaz*t(i,j))
           rspec=max(abs(un+c),abs(un-c))
           a_taup(:,:,i,j)=0.5_pr*(mat_a+rspec*mat_id)*l_tau(i,j)


           un=ux(i+1,j)*taux(i,j)+uy(i+1,j)*tauy(i,j)
           uij=ux(i+1,j); vij=uy(i+1,j)
           ec=(ux(i+1,j)**2+uy(i+1,j)**2)/2._pr
           h=(e(i+1,j)+p(i+1,j))/rho(i+1,j)

           mat_a(1,1)=0._pr; mat_a(1,2)=nx; mat_a(1,3)=ny; mat_a(1,4)=0._pr
           mat_a(2,1)=-uij*un+gamma1*ec*nx; mat_a(2,2)=un-gamma2*uij*nx; mat_a(2,3)=uij*ny-gamma1*vij*nx; mat_a(2,4)=gamma1*nx
           mat_a(3,1)=-vij*un+gamma1*ec*ny; mat_a(3,2)=vij*nx-gamma1*uij*ny; mat_a(3,3)=un-gamma2*vij*ny; mat_a(3,4)=gamma1*ny
           mat_a(4,1)=(gamma1*ec-h)*un; mat_a(4,2)=-gamma1*uij*un+h*nx; mat_a(4,3)=-gamma1*vij*un+h*ny; mat_a(4,4)=gamma*un

           c=sqrt(gamma*r_gaz*t(i+1,j))
           rspec=max(abs(un+c),abs(un-c))
           a_taum(:,:,i,j)=0.5_pr*(mat_a-rspec*mat_id)*l_tau(i,j)
        end do
     end do

     do j=0,jmax
        do i=1,imax
           un=ux(i,j)*nux(i,j)+uy(i,j)*nuy(i,j)
           uij=ux(i,j); vij=uy(i,j)
           nx=nux(i,j)
           ny=nuy(i,j)
           ec=(ux(i,j)**2+uy(i,j)**2)/2._pr
           h=(e(i,j)+p(i,j))/rho(i,j)

           mat_a(1,1)=0._pr; mat_a(1,2)=nx; mat_a(1,3)=ny; mat_a(1,4)=0._pr
           mat_a(2,1)=-uij*un+gamma1*ec*nx; mat_a(2,2)=un-gamma2*uij*nx; mat_a(2,3)=uij*ny-gamma1*vij*nx; mat_a(2,4)=gamma1*nx
           mat_a(3,1)=-vij*un+gamma1*ec*ny; mat_a(3,2)=vij*nx-gamma1*uij*ny; mat_a(3,3)=un-gamma2*vij*ny; mat_a(3,4)=gamma1*ny
           mat_a(4,1)=(gamma1*ec-h)*un; mat_a(4,2)=-gamma1*uij*un+h*nx; mat_a(4,3)=-gamma1*vij*un+h*ny; mat_a(4,4)=gamma*un

           c=sqrt(gamma*r_gaz*t(i,j))
           rspec=max(abs(un+c),abs(un-c))
           a_nup(:,:,i,j)=0.5_pr*(mat_a+rspec*mat_id)*l_nu(i,j)

           un=ux(i,j+1)*nux(i,j)+uy(i,j+1)*nuy(i,j)
           uij=ux(i,j+1); vij=uy(i,j+1)
           ec=(ux(i,j+1)**2+uy(i,j+1)**2)/2._pr
           h=(e(i,j+1)+p(i,j+1))/rho(i,j+1)

           mat_a(1,1)=0._pr; mat_a(1,2)=nx; mat_a(1,3)=ny; mat_a(1,4)=0._pr
           mat_a(2,1)=-uij*un+gamma1*ec*nx; mat_a(2,2)=un-gamma2*uij*nx; mat_a(2,3)=uij*ny-gamma1*vij*nx; mat_a(2,4)=gamma1*nx
           mat_a(3,1)=-vij*un+gamma1*ec*ny; mat_a(3,2)=vij*nx-gamma1*uij*ny; mat_a(3,3)=un-gamma2*vij*ny; mat_a(3,4)=gamma1*ny
           mat_a(4,1)=(gamma1*ec-h)*un; mat_a(4,2)=-gamma1*uij*un+h*nx; mat_a(4,3)=-gamma1*vij*un+h*ny; mat_a(4,4)=gamma*un

           c=sqrt(gamma*r_gaz*t(i,j+1))
           rspec=max(abs(un+c),abs(un-c))
           a_num(:,:,i,j)=0.5_pr*(mat_a-rspec*mat_id)*l_nu(i,j)
        end do
     end do

     do j=1,jmax
        do i=1,imax
           d(:,:,i,j)=mat_id+dt/aire(i,j)*(a_taup(:,:,i,j)-a_taum(:,:,i-1,j)+a_nup(:,:,i,j)-a_num(:,:,i,j-1))
        end do
     end do

     !--- solveur linaire : 
     !---1) descente
     do j=1,jmax
        do i=1,imax
           du1(:,i,j)=gauss(d(:,:,i,j),                               &
                &   -dt/aire(i,j)*(res(:,i,j)                              &
                &                   + (-matmul(a_taup(:,:,i-1,j),du1(:,i-1,j))             &
                &                     - matmul(a_nup(:,:,i,j-1),du1(:,i,j-1)))      &
                &                  )                                       &
                &          )
        end do
     end do

     !--- solveur linaire : 2) diagonale
     do j=1,jmax
        do i=1,imax
           du1(:,i,j)=matmul(d(:,:,i,j),du1(:,i,j))
        end do
     end do

     !--- solveur linaire : 3) remonte
     do j=jmax,1,-1
        do i=imax,1,-1
           du2(:,i,j)=gauss(d(:,:,i,j),                               &
                &   du1(:,i,j)-dt/aire(i,j)*(matmul(a_taum(:,:,i,j),du2(:,i+1,j))             &
                &                     + matmul(a_num(:,:,i,j),du2(:,i,j+1)))      &
                &          )
        end do
     end do

     !-- mise a jour
     vect_u=vect_u+du2(:,1:imax,1:jmax)

     !-- calcul de rho,ux,uy,t
     do i=1,imax
        do j=1,jmax
           rho(i,j)=vect_u(1,i,j)
           ux(i,j)=vect_u(2,i,j)/rho(i,j)
           uy(i,j)=vect_u(3,i,j)/rho(i,j)
           t(i,j)=2._pr/(3._pr*r_gaz*rho(i,j))*(vect_u(4,i,j)-0.5_pr*rho(i,j)*(ux(i,j)**2+uy(i,j)**2))
           if (t(i,j)<=0._pr) then
              print*,'t=',t(i,j),'en i,j=',i,j
              print*,'vect_u',vect_u(:,i,j)
              print*,'du2',du2(:,i,j)
              exit boucle_temps
           end if
        end do
     end do


     !-- mise a jour cl
     !-- exemple 1 : symtrie sur O et E, paroi solide au S, sortie sur partie N
     rho(0,1:jmax)=rho(1,1:jmax)      !- ouest (sortie)
     ux(0,1:jmax)=ux(1,1:jmax)
     uy(0,1:jmax)=uy(1,1:jmax)
     t(0,1:jmax)=t(1,1:jmax)

     rho(imax+1,1:jmax)=rho(imax,1:jmax)    !- est (sortie)
     ux(imax+1,1:jmax)=ux(imax,1:jmax)
     uy(imax+1,1:jmax)=uy(imax,1:jmax)
     t(imax+1,1:jmax)=t(imax,1:jmax)

     rho(1:imax,0)=rho(1:imax,1)          !- sud (paroi solide)
     t(1:imax,0)=t(1:imax,1)
     do i=1,imax
        un=nux(i,0)*ux(i,1)+nuy(i,0)*uy(i,1)
        ut=-nuy(i,0)*ux(i,1)+nux(i,0)*uy(i,1)
        ux(i,0)=nux(i,0)*(-un)-nuy(i,0)*ut
        uy(i,0)=nuy(i,0)*(-un)+nux(i,0)*ut
     end do

     !-- mise  jour des quantits drives (y compris dans les mailles fictives)
     p=rho*r_gaz*t
     b=sqrt(3*r_gaz*t)
     a=rho/(8_pr*b**3)
     e=0.5_pr*rho*(ux**2+uy**2)+3._pr/2._pr*rho*r_gaz*t

     n=n+1

  end do boucle_temps


  fich='SORTIES/resultats.vtk'
  open(unit=1,file=fich)
  write(1,'(1A26)') '# vtk DataFile Version 2.0'
  write(1,*) 'convection 2D'
  write(1,*) 'ASCII'
  write(1,*) 'DATASET STRUCTURED_GRID'
  write(1,*) 'DIMENSIONS ',imax+1,jmax+1,1
  write(1,*) 'POINTS',(imax+1)*(jmax+1),' double'
  do j=0,jmax
     do i=0,imax
        write(1,*) x(i,j),y(i,j),0.0_pr
     end do
  end do
  write(1,*) 'CELL_DATA',imax*jmax
  write(1,*) 'SCALARS rho double'
  write(1,*) 'LOOKUP_TABLE default'
  do j=1,jmax
     do i=1,imax
        write(1,*) rho(i,j)
     end do
  end do
  write(1,*) 'SCALARS T double'
  write(1,*) 'LOOKUP_TABLE default'
  do j=1,jmax
     do i=1,imax
        write(1,*) t(i,j)
     end do
  end do
  write(1,*) 'SCALARS p double'
  write(1,*) 'LOOKUP_TABLE default'
  do j=1,jmax
     do i=1,imax
        write(1,*) p(i,j)
     end do
  end do
  write(1,*) 'VECTORS u double'
  do j=1,jmax
     do i=1,imax
        write(1,*) ux(i,j),uy(i,j),0.0_pr
     end do
  end do
  close(1)

  !----------------------------------------------------------------

end program euler2d
