module mod_gauss
  
  implicit none

  contains

    function gauss(M,c) result(x)

      !-------------------- modules -------------------------!
      use mod_precision
      !------------------------------------------------------!

      implicit none

      !---------------------- in ----------------------------!
      real(pr), dimension(:), intent(in) :: c
      real(pr), dimension(:,:), intent(in) :: M
      !------------------------------------------------------!

      !---------------------- out ---------------------------!
      real(pr), dimension(size(c)) :: x
      !------------------------------------------------------!

      !-------------------- local var. ----------------------!
      integer :: n,i,j,k
      real(pr) :: pivot
      real(pr), dimension(size(c)) :: b
      real(pr), dimension(size(c),size(c)) :: A
      !------------------------------------------------------!

      n=size(c)
      A=M
      b=c
      !--- triangularisation
      do k=1,n-1
         pivot=a(k,k)
         if (pivot/=0._pr) then
            do i=k+1,n
               a(i,k+1:n)=a(i,k+1:n)-a(k,k+1:n)*a(i,k)/pivot
               b(i)=b(i)-b(k)*a(i,k)/pivot
            end do
         else
            print*,'function gauss : pivot nul'
            stop
         end if
      end do

      !---remonte
      if (a(n,n)/=0._pr) then
         x(n)=b(n)/a(n,n)
      end if
      do i=n-1,1,-1
         x(i)=(b(i)-sum(a(i,i+1:n)*x(i+1:n)))/a(i,i)
      end do


    end function gauss
    
    function gaussp(M,c) result(x)

      !-------------------- modules -------------------------!
      use mod_precision
      !------------------------------------------------------!

      implicit none

      !---------------------- in ----------------------------!
      real(pr), dimension(:), intent(in) :: c
      real(pr), dimension(:,:), intent(in) :: M
      !------------------------------------------------------!

      !---------------------- out ---------------------------!
      real(pr), dimension(size(c)) :: x
      !------------------------------------------------------!

      !-------------------- local var. ----------------------!
      integer :: n,i,j,k
      integer, dimension(1) :: p      
      real(pr) :: pivot
      real(pr) :: aux_b
      real(pr), dimension(size(c)) :: b,aux_a
      real(pr), dimension(size(c),size(c)) :: A
      !------------------------------------------------------!

      n=size(c)
      A=M
      b=c
      !--- triangularisation
      do k=1,n-1

         !-- change de lignes
         p=maxloc(abs(a(k:n,k)))+(k-1)
         if (p(1)/=k) then
            aux_a(k:n)=a(k,k:n)
            aux_b=b(k)
            
            a(k,k:n)=a(p(1),k:n)
            b(k)=b(p(1))

            a(p(1),k:n)=aux_a(k:n)
            b(p(1))=aux_b
         end if

         !-- triangularisation
         pivot=a(k,k)
         if (pivot/=0._pr) then
            do i=k+1,n
               a(i,k+1:n)=a(i,k+1:n)-a(k,k+1:n)*a(i,k)/pivot
               b(i)=b(i)-b(k)*a(i,k)/pivot
            end do
         else
            print*,'function gauss : pivot nul'
              stop
           end if

        end do

        !---remonte
        if (a(n,n)/=0._pr) then
           x(n)=b(n)/a(n,n)
        end if
        do i=n-1,1,-1
           x(i)=(b(i)-sum(a(i,i+1:n)*x(i+1:n)))/a(i,i)
        end do


      end function gaussp

    function gauss_mat(M,c) result(x)

      !-------------------- modules -------------------------!
      use mod_precision
      !------------------------------------------------------!

      implicit none

      !---------------------- in ----------------------------!
      real(pr), dimension(:,:), intent(in) :: c
      real(pr), dimension(:,:), intent(in) :: M
      !------------------------------------------------------!

      !---------------------- out ---------------------------!
      real(pr), dimension(size(c,1),size(c,2)) :: x
      !------------------------------------------------------!

      !-------------------- local var. ----------------------!
      integer :: n,i,j,k
      integer, dimension(1) :: p      
      real(pr) :: pivot
      real(pr), dimension(size(c,1)) :: aux_b,aux_a
      real(pr), dimension(size(c,1),size(c,2)) :: b
      real(pr), dimension(size(c,1),size(c,2)) :: A
      !------------------------------------------------------!

      n=size(c,1)
      A=M
      b=c
      !--- triangularisation
      do k=1,n-1

         !-- change de lignes
         p=maxloc(abs(a(k:n,k)))+(k-1)
         if (p(1)/=k) then
            aux_a(k:n)=a(k,k:n)
            aux_b=b(k,1:n)
            
            a(k,k:n)=a(p(1),k:n)
            b(k,1:n)=b(p(1),1:n)

            a(p(1),k:n)=aux_a(k:n)
            b(p(1),1:n)=aux_b
         end if

         !-- triangularisation
         pivot=a(k,k)
         if (pivot/=0._pr) then
            do i=k+1,n
               a(i,k+1:n)=a(i,k+1:n)-a(k,k+1:n)*a(i,k)/pivot
               b(i,1:n)=b(i,1:n)-b(k,1:n)*a(i,k)/pivot
            end do
         else
            print*,'function gauss : pivot nul'
              stop
           end if

        end do

        !---remonte
        if (a(n,n)/=0._pr) then
           x(n,1:n)=b(n,1:n)/a(n,n)
        end if
        do i=n-1,1,-1
           x(i,1:n)=(b(i,1:n)-matmul(a(i,i+1:n),x(i+1:n,1:n)))/a(i,i)
        end do


      end function gauss_mat


end module mod_gauss
